from os.path import abspath, dirname, join
from unittest import TestCase

import pytest

from iatikit import Sector
from iatikit.standard.codelist import Codelist
from iatikit.utils.exceptions import UnknownSectorVocabError, \
                                    UnknownSectorCodeError, \
                                    InvalidSectorCodeError
from iatikit.utils.config import CONFIG


class TestSector(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestSector, self).__init__(*args, **kwargs)
        standard_path = join(dirname(abspath(__file__)),
                             'fixtures', 'standard')
        config_dict = {'paths': {'standard': standard_path}}
        CONFIG.read_dict(config_dict)

    def test_sector_repr(self):
        sector = Sector(12345)
        assert str(sector) == '<Sector (12345, Vocabulary: Unspecified)>'

        sector = Sector(151, vocabulary=2)
        sector_repr = '<Sector (Government and civil society, general ' + \
                      '(151), Vocabulary: OECD DAC CRS Purpose Codes ' + \
                      '(3 digit))>'
        assert str(sector) == sector_repr

        sector = Sector('ABCD', vocabulary=3)
        sector_repr = '<Sector (ABCD, Vocabulary: Classification of the ' + \
                      'Functions of Government (UN))>'
        assert str(sector) == sector_repr

    def test_sector_percentage(self):
        sector = Sector(12345, percentage=100)
        assert isinstance(sector.percentage, float)
        assert sector.percentage == 100.0

    def test_sector_unknown_vocab(self):
        with pytest.raises(UnknownSectorVocabError):
            Sector(12345, vocabulary=10)

    def test_sector_unknown_code(self):
        with pytest.raises(UnknownSectorCodeError):
            Sector(12345, vocabulary=1)

    def test_sector_from_dac_5_codelist_item(self):
        codelist = Codelist('Sector', '2.03')
        codelist_item = codelist.get('73010')
        sector = Sector(codelist_item)
        assert sector.code.code == '73010'

    def test_sector_from_dac_3_codelist_item(self):
        codelist = Codelist('SectorCategory', '2.03')
        codelist_item = codelist.get('151')
        sector = Sector(codelist_item)
        assert sector.code.code == '151'

    def test_sector_from_bad_codelist_item(self):
        codelist = Codelist('Vocabulary', '1.05')
        codelist_item = codelist.get('DAC')
        with pytest.raises(InvalidSectorCodeError):
            Sector(codelist_item)



from .activity_schema import get_activity_schema
# from .organisation_schema import get_organisation_schema


def get_schema(filetype, version):
    if filetype == 'activity':
        return get_activity_schema(version)
    raise NotImplementedError()
    # elif filetype == 'organisation':
    #     return get_organisation_schema(version)



from .standard.codelist import codelists  # noqa: F401
from .data.registry import Registry
from .data.publisher import Publisher  # noqa: F401
from .data.dataset import Dataset  # noqa: F401
from .data.activity import Activity  # noqa: F401
from .data.sector import Sector  # noqa: F401
from .utils import download  # noqa: F401
from .utils.config import CONFIG  # noqa: F401
from .__version__ import __version__  # noqa: F401


def data():
    """Helper function for constructing a Registry object."""
    return Registry()



from os.path import basename, exists, splitext
from glob import glob
import json
import logging
import webbrowser

from past.builtins import basestring
from lxml import etree as ET

from ..utils.abstract import GenericSet
from ..utils.exceptions import SchemaNotFoundError, MappingsNotFoundError
from ..utils.validator import Validator, ValidationError
from ..standard.xsd_schema import XSDSchema
from ..standard.codelist_mappings import CodelistMappings
from .activity import ActivitySet


class Dataset(object):
    """Class representing an IATI dataset."""

    def __init__(self, data_path, metadata_path=None):
        """Construct a new Dataset object.

        The file locations of the data and metadata must be specified with
        the ``data_path`` and ``metadata_path`` arguments.
        """
        self.data_path = data_path
        self.metadata_path = metadata_path
        self._etree = None
        self._metadata = None
        self._schema = None

    @property
    def name(self):
        """Return the name of this dataset, derived from the filename."""
        if isinstance(self.data_path, basestring):
            return splitext(basename(self.data_path))[0]
        elif isinstance(self.metadata_path, basestring):
            return splitext(basename(self.metadata_path))[0]
        else:
            return 'dataset'

    @property
    def etree(self):
        """Return the XML of this dataset, as an lxml element tree."""
        if not self._etree:
            try:
                parser = ET.XMLParser(remove_blank_text=True)
                self._etree = ET.parse(self.data_path, parser)
            except ET.XMLSyntaxError:
                logging.warning('Dataset "%s" XML is invalid', self.name)
                raise
        return self._etree

    @property
    def xml(self):
        """Return the raw XML of this dataset, as a byte-string."""
        return bytes(ET.tostring(self.etree, pretty_print=True))

    def __repr__(self):
        return '<{} ({})>'.format(self.__class__.__name__, self.name)

    def show(self):
        """Open a new browser tab to the iatiregistry.org page
        for this dataset.
        """
        name = self.metadata.get('name')
        if name:
            url = 'https://iatiregistry.org/dataset/{}'.format(name)
            webbrowser.open_new_tab(url)
            return True
        logging.warning('Can\'t show dataset - metadata missing.')
        return False

    def _get_schema(self):
        """Get the XSD Schema for this dataset. Raise exception on error."""
        if not self._schema:
            self._schema = XSDSchema(self.filetype, self.version)
        return self._schema

    @property
    def schema(self):
        """Get the XSD Schema for this dataset."""
        try:
            return self._get_schema()
        except SchemaNotFoundError as error:
            logging.warning(str(error))

    def unminify_xml(self):
        self._etree = ET.ElementTree(ET.fromstring(self.xml))

    def validate_xml(self):
        """Check whether the XML in this dataset can be parsed."""
        try:
            self.etree
        except ET.XMLSyntaxError as error:
            return Validator(False, [ValidationError(str(error))])
        return Validator(True)

    def validate_iati(self):
        """Validate dataset against the relevant IATI schema."""
        xml_valid = self.validate_xml()
        if not xml_valid:
            msg = 'Can\'t perform IATI schema validation for ' + \
                  'invalid XML.'
            return Validator(False, [ValidationError(msg)])
        try:
            return self._get_schema().validate(self.etree)
        except SchemaNotFoundError as error:
            logging.warning(str(error))
            return Validator(False, [ValidationError(str(error))])

    def validate_codelists(self):
        """Validate dataset against the relevant IATI codelists."""
        xml_valid = self.validate_xml()
        if not xml_valid:
            msg = 'Can\'t perform codelist validation for ' + \
                  'invalid XML.'
            return Validator(False, [ValidationError(msg)])
        try:
            mappings = CodelistMappings(self.filetype, self.version)
        except MappingsNotFoundError:
            msg = 'Can\'t perform codelist validation for ' + \
                  'IATI version %s datasets.'
            logging.warning(msg, self.version)
            return Validator(True)
        return mappings.validate(self)

    @property
    def metadata(self):
        """Return a dictionary of registry metadata for this dataset."""
        if self._metadata is None:
            if self.metadata_path is not None and exists(self.metadata_path):
                with open(self.metadata_path) as handler:
                    self._metadata = json.load(handler)
                extras = self.metadata.get('extras')
                self._metadata['extras'] = {x['key']: x['value']
                                            for x in extras}
            else:
                msg = 'No metadata was found for dataset "%s"'
                logging.warning(msg, self.name)
                self._metadata = {}
        return self._metadata

    @property
    def filetype(self):
        """Return the filetype according to the metadata
        (i.e. "activity" or "organisation").

        If it can't be found in the metadata, revert to using
        the XML root node.

        Returns None if the filetype can't be determined.
        """
        try:
            filetype = self.metadata['extras']['filetype']
            if filetype in ['activity', 'organisation']:
                return filetype
        except KeyError:
            pass

        try:
            return {
                'iati-activities': 'activity',
                'iati-organisations': 'organisation',
            }[self.root]
        except KeyError:
            pass

    @property
    def root(self):
        """Return the name of the XML root node."""
        return self.etree.getroot().tag

    @property
    def version(self):
        """Return the IATI version according to the XML root node.

        Return "1.01" if the version can't be determined.
        """
        version = self.etree.getroot().get('version')
        if version is not None:
            return version

        logging.warning('@version attribute is not declared. Assuming "1.01".')
        # default version
        return '1.01'

    @property
    def activities(self):
        """Return an iterator of all activities in this dataset."""
        return ActivitySet([self])


class DatasetSet(GenericSet):
    """Class representing a grouping of ``Dataset`` objects.

    Objects in this grouping can be filtered and iterated over.
    Queries are only constructed and run when needed, so they
    can be efficient.
    """

    _key = 'name'
    _filters = ['name', 'filetype']
    _instance_class = Dataset

    def __init__(self, data_path, metadata_path, **kwargs):
        super(DatasetSet, self).__init__()
        self.wheres = kwargs

        self.data_path = data_path
        self.metadata_path = metadata_path

    def __iter__(self):
        data_paths = {splitext(basename(x))[0]: x
                      for x in glob(self.data_path)}
        metadata_paths = {splitext(basename(x))[0]: x
                          for x in glob(self.metadata_path)}
        paths = {x: (data_paths.get(x), metadata_paths.get(x))
                 for x in set(list(data_paths.keys()) +
                              list(metadata_paths.keys()))}

        name = self.wheres.get('name')
        if name is not None:
            paths = [paths[name]] if name in paths else []
        else:
            paths = sorted(list(paths.values()))

        where_filetype = self.wheres.get('filetype')

        for data_path, metadata_path in paths:
            dataset = Dataset(data_path, metadata_path)
            if where_filetype is not None and \
                    dataset.filetype != where_filetype:
                continue
            yield dataset



from copy import copy
from itertools import islice
from .exceptions import FilterError


class GenericSet(object):
    """Class representing a generic grouping of iatikit objects.

    Objects in this grouping can be filtered and iterated over.
    Queries are only constructed and run when needed, so they
    can be efficient.
    """

    _key = None
    _filters = []
    _instance_class = None

    def __init__(self):
        self.wheres = {}

    def where(self, **kwargs):
        """Return a new set, with the filters provided in ``**kwargs``.
        """
        for k in kwargs:
            if k.split('__')[0] not in self._filters:
                raise FilterError('Unknown filter: {}'.format(k))
        clone = copy(self)
        clone.wheres = dict(clone.wheres, **kwargs)
        return clone

    def filter(self, **kwargs):
        """Return a new set, with the filters provided in ``**kwargs``.

        Alias of ``where(**kwargs)``.
        """
        return self.where(**kwargs)

    def __getitem__(self, index):
        try:
            return next(islice(self, index, index + 1))
        except TypeError:
            return list(islice(self, index.start, index.stop, index.step))
        except StopIteration:
            pass
        raise IndexError('index out of range')

    def __len__(self):
        return sum(1 for x in self)

    def count(self):
        """The number of items in this set.

        Equivalent to ``len(self)``.
        """
        return len(self)

    def first(self):
        """Return the first item in this set.

        Raises an ``IndexError`` if the set contains zero items.

        Equivalent to ``self[0]``.
        """
        return self[0]

    def all(self):
        """Return a list of all items in this set.

        Equivalent to ``list(self)``.
        """
        return list(self)

    def get(self, item, default=None, fast=False):
        """Return an item from the set, according to the primary key.

        If no matching item is found, ``default`` is returned.
        """
        if fast:
            msg = 'The `fast` keyword argument can\'t be used here.'
            raise NotImplementedError(msg)
        if isinstance(item, self._instance_class):
            item = getattr(item, self._key)
        try:
            return self.find(**{self._key: item})
        except IndexError:
            return default

    def find(self, **kwargs):
        """Return the first matching item from the set, according to the
        filters provided in ``kwargs``.

        If no matching item is found, an ``IndexError`` is raised.

        ``find(**kwargs)`` is equivalent to ``where(**kwargs).first()``.
        """
        return self.where(**kwargs).first()


class GenericType(object):
    def __init__(self, expr):
        self._expr = expr

    def get(self):
        return self._expr

    def run(self, etree):
        return etree.xpath(self.get())

    def where(self, operation, value):
        if operation == 'exists':
            sub_operation = '!= 0' if value else '= 0'
            return 'count({expr}) {subop}'.format(
                expr=self.get(),
                subop=sub_operation,
            )
        elif operation == 'eq':
            return '{expr} = "{value}"'.format(
                expr=self.get(),
                value=value,
            )
        raise FilterError('Unknown filter modifier: {}'.format(operation))



from datetime import datetime
from os.path import exists, join
import json
import warnings

from .publisher import PublisherSet
from .dataset import DatasetSet
from .activity import ActivitySet
from ..utils.exceptions import NoDataError
from ..utils.config import CONFIG


class Registry(object):
    """Class representing the IATI registry."""

    def __init__(self):
        """Construct a new Registry object.

        A warning is raised if the data is more than 7 days old.

        A ``NoDataError`` is raised if there is no data.
        """
        self._last_updated = None
        self.path = CONFIG['paths']['registry']

        last_updated = self.last_updated
        days_ago = (datetime.now() - last_updated).days
        if days_ago > 7:
            warning_msg = 'Warning: Data was last updated {} days ' + \
                          'ago. Consider downloading a fresh ' + \
                          'data dump, using:\n\n   ' + \
                          '>>> iatikit.download.data()\n'
            warnings.warn(warning_msg.format(days_ago))

    @property
    def last_updated(self):
        """Return the datetime when the local cache was last updated.
        """
        if not self._last_updated:
            filepath = join(self.path, 'metadata.json')
            if exists(filepath):
                with open(filepath) as handler:
                    j = json.load(handler)
                last_updated = j['updated_at']
                self._last_updated = datetime.strptime(
                    last_updated, '%Y-%m-%dT%H:%M:%SZ')
            else:
                error_msg = 'Error: No data found! ' + \
                            'Download a fresh data dump ' + \
                            'using:\n\n   ' + \
                            '>>> iatikit.download.data()\n'
                raise NoDataError(error_msg)
        return self._last_updated

    @property
    def publishers(self):
        """Return an iterator of all publishers on the registry."""
        data_path = join(self.path, 'data', '*')
        metadata_path = join(self.path, 'metadata', '*')
        return PublisherSet(data_path, metadata_path)

    @property
    def datasets(self):
        """Return an iterator of all IATI datasets on the registry."""
        publisher_set = self.publishers
        data_path = join(publisher_set.data_path, '*')
        metadata_path = join(publisher_set.metadata_path, '*')
        return DatasetSet(data_path, metadata_path)

    @property
    def activities(self):
        """Return an iterator of all IATI activities on the registry."""
        return ActivitySet(self.datasets)



import json
import logging
from os.path import basename, exists, join, split, splitext
from glob import glob
import webbrowser

from past.builtins import basestring

from ..utils.abstract import GenericSet
from .dataset import DatasetSet
from .activity import ActivitySet


class Publisher(object):
    """Class representing an IATI publisher."""

    def __init__(self, data_path, metadata_path, metadata_filepath):
        self.data_path = data_path
        self.metadata_path = metadata_path
        self.metadata_filepath = metadata_filepath
        self._metadata = None

    @property
    def name(self):
        """Return the "registry name" or "shortname" of this publisher,
        derived from the filepath.
        """
        if isinstance(self.data_path, basestring):
            return split(self.data_path)[1]
        elif isinstance(self.metadata_path, basestring):
            return split(self.metadata_path)[1]
        else:
            return 'publisher'

    def __repr__(self):
        return '<{} ({})>'.format(self.__class__.__name__, self.name)

    def show(self):
        """Open a new browser tab to the iatiregistry.org page
        for this publisher.
        """
        name = self.metadata.get('name')
        if name:
            url = 'https://iatiregistry.org/publisher/{}'.format(name)
            webbrowser.open_new_tab(url)
            return True
        logging.warning('Can\'t show publisher - metadata missing.')
        return False

    @property
    def datasets(self):
        """Return an iterator of all datasets for this publisher."""
        data_path = join(self.data_path, '*')
        metadata_path = join(self.metadata_path, '*')
        return DatasetSet(data_path, metadata_path)

    @property
    def activities(self):
        """Return an iterator of all activities for this publisher."""
        return ActivitySet(self.datasets)

    @property
    def metadata(self):
        """Return a dictionary of registry metadata for this publisher."""
        if self._metadata is None:
            if exists(self.metadata_filepath):
                with open(self.metadata_filepath) as handler:
                    self._metadata = json.load(handler)
            else:
                msg = 'No metadata was found for publisher "%s"'
                logging.warning(msg, self.name)
                self._metadata = {}
        return self._metadata


class PublisherSet(GenericSet):
    """Class representing a grouping of ``Publisher`` objects.

    Objects in this grouping can be filtered and iterated over.
    Queries are only constructed and run when needed, so they
    can be efficient.
    """

    _filters = ['name']
    _key = 'name'
    _instance_class = Publisher

    def __init__(self, data_path, metadata_path, **kwargs):
        super(PublisherSet, self).__init__()
        self.wheres = kwargs
        self.data_path = data_path
        self.metadata_path = metadata_path

    def __iter__(self):
        data_paths = {basename(x): x
                      for x in glob(self.data_path)
                      if not x.endswith('.json')}
        metadata_paths = {basename(x): x
                          for x in glob(self.metadata_path)
                          if not x.endswith('.json')}
        metadata_filepaths = {splitext(basename(x))[0]: x
                              for x in glob(self.metadata_path + '.json')}
        paths = {x: (data_paths.get(x),
                     metadata_paths.get(x),
                     metadata_filepaths.get(x))
                 for x in set(list(data_paths.keys()) +
                              list(metadata_paths.keys()) +
                              list(metadata_filepaths.keys()))}

        name = self.wheres.get('name')
        if name is not None:
            paths = [paths[name]] if name in paths else []
        else:
            paths = sorted(list(paths.values()))

        for data_path, metadata_path, metadata_filepath in paths:
            yield Publisher(data_path, metadata_path, metadata_filepath)



# from ..utils.exceptions import SchemaError
# from ..utils.types import StringType


# # pylint: disable=too-many-ancestors
# class OrganisationSchema101(object):
#     version = '1.01'

#     @classmethod
#     def org_identifier(cls):
#         return StringType('iati-identifier/text()')


# class OrganisationSchema102(OrganisationSchema101):
#     version = '1.02'


# class OrganisationSchema103(OrganisationSchema102):
#     version = '1.03'


# class OrganisationSchema104(OrganisationSchema103):
#     version = '1.04'


# class OrganisationSchema105(OrganisationSchema104):
#     version = '1.05'


# class OrganisationSchema201(OrganisationSchema105):
#     version = '2.01'

#     @classmethod
#     def org_identifier(cls):
#         return StringType('organisation-identifier/text()')


# class OrganisationSchema202(OrganisationSchema201):
#     version = '2.02'


# class OrganisationSchema203(OrganisationSchema202):
#     version = '2.03'


# def get_organisation_schema(version):
#     schema = None

#     if version == '2.03':
#         schema = OrganisationSchema203
#     elif version == '2.02':
#         schema = OrganisationSchema202
#     elif version == '2.01':
#         schema = OrganisationSchema201
#     elif version == '1.05':
#         schema = OrganisationSchema105
#     elif version == '1.04':
#         schema = OrganisationSchema104
#     elif version == '1.03':
#         schema = OrganisationSchema103
#     elif version == '1.02':
#         schema = OrganisationSchema102
#     elif version == '1.01':
#         schema = OrganisationSchema101

#     if not schema:
#         msg = 'Unknown organisation schema: version: {}'.format(version)
#         raise SchemaError(msg)

#     return schema



import datetime
from os.path import abspath, dirname, join
from unittest import TestCase

from mock import patch
from lxml import etree as ET
import pytest

from iatikit.data.dataset import DatasetSet, Dataset
from iatikit.data.activity import ActivitySet, Activity
from iatikit.standard.activity_schema import ActivitySchema105
from iatikit.utils.config import CONFIG
from iatikit import Sector


class TestActivitySet(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestActivitySet, self).__init__(*args, **kwargs)
        registry_path = join(dirname(abspath(__file__)),
                             'fixtures', 'registry')
        org_datasets = DatasetSet(
            join(registry_path, 'data', 'fixture-org', '*'),
            join(registry_path, 'metadata', 'fixture-org', '*'),
        )
        self.fixture_org_acts = ActivitySet(org_datasets)

        standard_path = join(dirname(abspath(__file__)), 'fixtures',
                             'standard')
        config_dict = {'paths': {'standard': standard_path}}
        CONFIG.read_dict(config_dict)

    def test_activities_iter(self):
        activities = list(self.fixture_org_acts)
        assert len(activities) == 4

    def test_activities_len(self):
        assert len(self.fixture_org_acts) == 4

    def test_activities_filter_by_id(self):
        iati_id = 'GB-COH-01234567-Humanitarian Aid-1'
        acts = self.fixture_org_acts.where(id=iati_id).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == iati_id

    def test_activities_filter_by_iati_identifier(self):
        iati_id = 'GB-COH-01234567-Humanitarian Aid-1'
        acts = self.fixture_org_acts.where(iati_identifier=iati_id).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == iati_id

    def test_activities_fast_get(self):
        iati_id = 'GB-COH-01234567-Humanitarian Aid-1'
        act = self.fixture_org_acts.get(iati_id, fast=True)
        assert act.iati_identifier == iati_id

    def test_activities_filter_by_title_v1(self):
        title = 'Humanitarian Aid - Implementer 2'
        acts = self.fixture_org_acts.where(title=title).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-Humanitarian Aid-1'

    def test_activities_filter_by_title_startswith(self):
        title_start = 'Development'
        acts = self.fixture_org_acts.where(title__startswith=title_start).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-1'

    def test_activities_filter_by_title_v2(self):
        title = 'Development work'
        acts = self.fixture_org_acts.where(title=title).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-1'

    def test_activities_filter_by_description(self):
        description = 'Just a short description.'
        acts = self.fixture_org_acts.where(
            description=description).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-Humanitarian Aid-1'

    def test_activities_filter_by_sector(self):
        sector = Sector('15163', vocabulary='DAC')
        acts = self.fixture_org_acts.where(
            sector=sector).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-1'

    def test_activities_filter_by_bad_sector(self):
        err_msg = 'bad-sector is not a sector'
        with pytest.raises(Exception, match=err_msg):
            self.fixture_org_acts.where(sector='bad-sector').all()

    # TODO
    def test_activities_filter_by_sector_in(self):
        sector = Sector('151', vocabulary='2')
        acts = self.fixture_org_acts.where(
            sector__in=sector).all()
        assert len(acts) == 1

    def test_activities_filter_by_bad_sector_in(self):
        err_msg = 'bad-sector-cat is not a sector category'
        with pytest.raises(Exception, match=err_msg):
            self.fixture_org_acts.where(sector__in='bad-sector-cat').all()

    def test_activities_filter_by_sector_exists(self):
        acts = self.fixture_org_acts.where(
            sector__exists=False).all()
        assert len(acts) == 1

    def test_activities_filter_by_planned_start_v2(self):
        acts = self.fixture_org_acts.where(
            planned_start='2011-11-01').all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-1'

    def test_activities_filter_by_actual_start_v2(self):
        acts = self.fixture_org_acts.where(
            actual_start='2011-10-01').all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-1'

    def test_activities_filter_by_planned_end_v2(self):
        planned_end = datetime.date(2019, 9, 30)
        acts = self.fixture_org_acts.where(
            planned_end=planned_end).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-1'

    def test_activities_filter_by_actual_end_exists(self):
        acts = self.fixture_org_acts.where(
            actual_end__exists=True).all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-Humanitarian Aid-1'

    def test_activities_filter_by_xpath(self):
        acts = self.fixture_org_acts.where(
            xpath='activity-status/@code="2"').all()
        assert len(acts) == 1
        assert acts[0].iati_identifier == 'GB-COH-01234567-Humanitarian Aid-0'


class TestActivity(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestActivity, self).__init__(*args, **kwargs)
        data_path = join(dirname(abspath(__file__)),
                         'fixtures', 'registry', 'data', 'fixture-org',
                         'fixture-org-activities2.xml')
        etree = ET.parse(data_path)
        activity_el = etree.xpath('//iati-activity')[0]
        activity_el1 = etree.xpath('//iati-activity')[1]
        activity_el2 = etree.xpath('//iati-activity')[2]
        dataset = Dataset(data_path, None)
        schema = ActivitySchema105()
        self.activity = Activity(activity_el, dataset, schema)
        self.activity1 = Activity(activity_el1, dataset, schema)
        self.activity2 = Activity(activity_el2, dataset, schema)

    @patch('webbrowser.open_new_tab')
    def test_activity_show(self, fake_open_new_tab):
        url = 'http://d-portal.org/q.html' + \
              '?aid=GB-COH-01234567-Humanitarian+Aid-0'
        self.activity.show()
        fake_open_new_tab.assert_called_once_with(url)

    def test_activity_xml(self):
        xml = self.activity.xml
        first_line = '<iati-activity default-currency="GBP" ' + \
                     'last-updated-datetime="2016-01-12T15:17:15">'
        assert xml.decode().split('\n', 1)[0] == first_line

    def test_activity_id(self):
        id_ = 'GB-COH-01234567-Humanitarian Aid-0'
        assert self.activity.id == id_
        assert self.activity.iati_identifier == id_

    def test_activity_repr(self):
        act_repr = '<Activity (GB-COH-01234567-Humanitarian Aid-0)>'
        assert str(self.activity) == act_repr

    def test_activity_title(self):
        title = 'Humanitarian Aid - Implementor 1'
        assert self.activity.title[0] == title

    def test_activity_description(self):
        assert self.activity.description == []

    # TODO
    def test_activity_sector(self):
        sector = Sector('73010', vocabulary='1')
        assert self.activity.sector[0] == sector

    def test_activity_planned_start(self):
        assert self.activity2.planned_start == datetime.date(2009, 1, 1)

    def test_activity_start(self):
        assert self.activity.start == datetime.date(2013, 4, 16)

    def test_activity_no_end(self):
        assert self.activity.end is None

    def test_activity_end(self):
        assert self.activity1.end == datetime.date(2015, 1, 16)



import json
from os.path import exists, join

from ..utils.abstract import GenericSet
from ..utils.exceptions import NoCodelistsError
from ..utils.config import CONFIG


class CodelistItem(object):
    def __init__(self, codelist, **kwargs):
        self.category = kwargs.get('category')
        self.status = kwargs.get('status', 'active')
        self.code = kwargs.get('code')
        self.name = kwargs.get('name')
        self.description = kwargs.get('description')
        self.codelist = codelist

    def __repr__(self):
        return '<{} ({} ({}))>'.format(
            self.__class__.__name__,
            self.name,
            self.code)

    def __eq__(self, value):
        return self.code == value.code and \
               self.codelist.slug == value.codelist.slug

    def __ne__(self, value):
        return not self.__eq__(value)


class Codelist(GenericSet):
    _key = 'code'
    _filters = ['code', 'name', 'version', 'category']
    _instance_class = CodelistItem

    def __init__(self, slug, version, **kwargs):
        super(Codelist, self).__init__()
        self.wheres = kwargs
        self.slug = slug
        self.path = join(CONFIG['paths']['standard'],
                         'codelists', slug + '.json')
        self.version = version
        self.__data = None

    @property
    def _data(self):
        if not self.__data:
            with open(self.path) as handler:
                self.__data = json.load(handler)
        return self.__data

    @property
    def data(self):
        return self._data['data']

    @property
    def attributes(self):
        return self._data['attributes']

    @property
    def metadata(self):
        return self._data['metadata']

    def __iter__(self):
        code = self.wheres.get('code')
        name = self.wheres.get('name')
        category = self.wheres.get('category')
        version = self.wheres.get('version', self.version)
        if version is not None:
            version = str(version)
        if code is not None:
            code = str(code)
        if category is not None:
            category = str(category)
        for data in self.data.values():
            if code is not None and data['code'] != code:
                continue
            if name is not None and data['name'] != name:
                continue
            if category is not None and data['category'] != category:
                continue
            if version is not None:
                version_from = data.get('from')
                version_until = data.get('until')
                if version_from and version_until and \
                        (version < version_from or
                         version > version_until):
                    continue
            yield CodelistItem(self, **data)

    def __repr__(self):
        if self.version:
            slug = '{} v{}'.format(self.slug, self.version)
        else:
            slug = self.slug
        return '<{} ({})>'.format(self.__class__.__name__, slug)

    @property
    def url(self):
        return self.metadata.get('url')

    @property
    def name(self):
        return self.metadata.get('name')

    @property
    def description(self):
        return self.metadata.get('description')

    @property
    def complete(self):
        return self.attributes.get('complete') == '1'


class CodelistSet(GenericSet):
    _key = 'slug'
    _filters = ['slug', 'version']
    _instance_class = Codelist

    def __init__(self, **kwargs):
        super(CodelistSet, self).__init__()
        self.wheres = kwargs
        self.path = join(CONFIG['paths']['standard'], 'codelists')
        if not exists(join(self.path, 'codelists.json')):
            error_msg = 'Error: No codelists found! ' + \
                          'Download fresh codelists ' + \
                          'using:\n\n   ' + \
                          '>>> iatikit.download.codelists()\n'
            raise NoCodelistsError(error_msg)

    def __iter__(self):
        version = self.wheres.get('version')
        if version:
            version = str(version)
        slug = self.wheres.get('slug')
        with open(join(self.path, 'codelists.json')) as handler:
            all_codelists = json.load(handler)
        for codelist_slug, codelist_versions in all_codelists.items():
            if version is not None and version not in codelist_versions:
                continue

            if slug is not None and slug != codelist_slug:
                continue

            yield Codelist(codelist_slug, version)


def codelists():
    """Helper function for fetching all codelists."""
    return CodelistSet()



import json
from os.path import exists, join

from ..utils.exceptions import MappingsNotFoundError
from ..utils.validator import Validator, ValidationError
from ..utils.config import CONFIG
from .codelist import CodelistSet


class CodelistValidationError(ValidationError):
    def __init__(self, msg, line, path, codelist, version):
        super(CodelistValidationError, self).__init__(msg, line, None, path)

        details = 'Only values from the {codelist_name} ' + \
                  'codelist are permitted.'
        self.details = details.format(codelist_name=codelist.name)
        self.codelist_slug = codelist.slug
        self.version = version

    @property
    def url(self):
        if self.version in ['1.01', '1.02', '1.03']:
            slug = self.codelist_slug.lower().replace(' ', '_')
        else:
            slug = self.codelist_slug

        version_str = self.version.replace('.', '')
        tmpl = 'http://reference.iatistandard.org/{version}/codelists/{slug}/'
        return tmpl.format(version=version_str, slug=slug)


class CodelistMappings(object):
    def __init__(self, filetype, version):
        self.filetype = filetype
        self.version = version

        version_path = version.replace('.', '')
        self.mappings_path = join(CONFIG['paths']['standard'],
                                  'codelist_mappings', version_path,
                                  '{}-mappings.json'.format(filetype))

        if not exists(self.mappings_path):
            tmpl = 'No codelist mappings found for IATI version ' + \
                   '"{version} ({filetype})".'
            msg = tmpl.format(version=version, filetype=filetype)
            raise MappingsNotFoundError(msg)

    def __repr__(self):
        return '<{} ({} {})>'.format(self.__class__.__name__,
                                     self.filetype, self.version)

    def validate(self, dataset):
        codelists = CodelistSet(version=dataset.version)

        def parse_mapping(mapping):
            condition = mapping.get('condition')
            if condition:
                path_body, path_head = mapping['path'].rsplit('/', 1)
                path = '{path_body}[{condition}]/{path_head}'.format(
                    path_body=path_body,
                    condition=condition,
                    path_head=path_head,
                )
            else:
                path = mapping['path']
            return path, codelists.get(mapping['codelist'])

        with open(self.mappings_path) as handler:
            mappings = json.load(handler)

        def get_path(value):
            if value.is_text:
                output = ['text()']
            elif value.is_attribute:
                attr_name = value.attrname
                output = ['@' + attr_name]
            else:
                raise Exception('Weird')
            while True:
                value = value.getparent()
                if value is None:
                    break
                tag = value.tag
                idx = len(list(value.itersiblings(tag, preceding=True))) + 1
                output.append('{tag}[{idx}]'.format(tag=tag, idx=idx))
            return '//{}'.format('/'.join(output[::-1]))

        success = True
        error_log = []
        for mapping in mappings:
            xpath_query, codelist = parse_mapping(mapping)
            values = dataset.etree.xpath(xpath_query)
            for value in set(values):
                if codelist.get(value):
                    continue
                line = value.getparent().sourceline
                path = get_path(value)
                msg = 'The value "{}" is not in the {} codelist.'.format(
                    value, codelist.name)
                codelist_error = CodelistValidationError(
                    msg, line, path, codelist, self.version)
                error_log.append(codelist_error)
                success = False
        return Validator(success, error_log)



import requests


def get_iati_versions():
    versions_url = 'http://reference.iatistandard.org/201/codelists/' + \
                   'downloads/clv2/json/en/Version.json'
    versions = [d['code']
                for d in requests.get(versions_url).json()['data']]
    versions.reverse()
    return versions
